package moreswing.swing

import scala.collection.Iterator
import scala.collection.mutable.Buffer

/** Default partial implementation for buffer adapters.
 *
 * @author myst3r10n
 */
protected[moreswing] abstract class BufferWrapper[T] extends Buffer[T] { outer =>

  /** Clear buffer. */
  def clear() { for (i <- 0 until length) remove(0) }

  /** Update buffer.
   *
   * @param n position
   * @param element to update
   */
  def update(n: Int, element: T) {

    remove(n)
    insertAt(n, element)

  }

  /** Inserts elements of other buffer into this buffer.
   *
   * @param n position
   * @param other buffer wrapper to insert
   */
  def insertAll(n: Int, other: Traversable[T]) {
    var i = n
    for (el <- other) {
      insertAt(i, el)
      i += 1
    }
  }

  /** Insert element into this buffer wrapper.
   *
   * @param n position
   * @param element to insert
   */
  protected def insertAt(n: Int, element: T)

  /** Append element into buffer wrapper.
   *
   * @param element to append
   *
   * @return itself
   */
  def +=:(element: T): this.type = { insertAt(0, element); this }

  /** Creates a new iterator over all elements contained in this iterable object. */
  def iterator = Iterator.range(0,length).map(apply(_))

}

